# Figure 9.2
from random import choice

terminals = {'a', 'the', 'hungry', 'sleepy', 'cat', 'dog',
             'chases', 'runs', 'slowly', 'quickly'}

productions = {
  'SENTENCE': [['NOUN_PHRASE', 'VERB_PHRASE', 'NOUN_PHRASE'],
               ['NOUN_PHRASE', 'VERB_PHRASE']],
  'NOUN_PHRASE': [['ARTICLE', 'ADJECTIVE', 'NOUN'],
                  ['ARTICLE', 'NOUN']],
  'VERB_PHRASE': [['VERB', 'ADVERB'], ['VERB']],
  'ARTICLE': ['a', 'the'],
  'ADJECTIVE': ['hungry', 'sleepy'],
  'NOUN': ['cat', 'dog'],
  'VERB': ['chases', 'runs'],
  'ADVERB': ['slowly', 'quickly']}

nonterminals = set(productions.keys())

G = {'nonterminals': nonterminals,
     'terminals': terminals,
     'start_symbol': 'SENTENCE',
     'productions': productions}

def nonterminals_remain(grammar, segment):
  nonterminals = grammar['nonterminals']
  return any(map(lambda s: s in nonterminals, segment))

def derive(grammar, LHS):
  RHS = grammar['productions'][LHS]
  if isinstance(RHS[0], list):
    return choice(RHS)
  else:
    return [choice(RHS)]

def find_derivation(grammar):
  derivation = [grammar['start_symbol']]
  terminals = grammar['terminals']
  while nonterminals_remain(grammar, derivation):
    print(' '.join(derivation))
    derive_one = []
    for i in range(len(derivation)):
      symbol = derivation[i]
      if symbol in terminals:
        derive_one.append(symbol)
      else:
        derive_one.extend(derive(grammar, symbol))
        derive_one.extend(derivation[i+1:])
        break
    derivation = derive_one
  return ' '.join(derivation)

print(find_derivation(G))

# Code for Exercise 371
"""Generate random sentences from a grammar.  The grammar
consists of entries that can be written as S = 'NP VP | S and S',
which gets translated to {'S': [['NP', 'VP'], ['S', 'and', 'S']]}, and
means that one of the top-level lists will be chosen at random, and
then each element of the second-level list will be rewritten; if a symbol is
not in the grammar it rewrites as itself.   The functions generate and
generate_tree generate a string and tree representation, respectively, of
a random sentence."""

import random

def Grammar(**grammar):
  "Create a dictionary mapping symbols to alternatives."
  for (cat, rhs) in grammar.items():
    grammar[cat] = [alt.split() for alt in rhs.split('|')]
  return grammar

grammar = Grammar(
  S   = 'NP VP | S and S',
  NP  = 'Art N | Name',
  VP  = 'V NP',
  Art = 'the | a | every | some',
  N   = 'man | ball | woman | table | dog | cat | wombat',
  V   = 'hit | took | saw | liked | worshiped | remembered',
  Name= 'Alice | Bob | Carlos | Dan | Eve'
  )

def generate(symbol='S'):
  "Replace symbol with a random entry in grammar (recursively); join into a string."
  if symbol not in grammar:
    return symbol
  else:
    return ' '.join(map(generate, random.choice(grammar[symbol])))

def generate_tree(symbol='S'):
  "Replace symbol with a random entry in grammar (recursively); return a tree."
  if symbol not in grammar:
    return symbol
  else:
    return {symbol: [generate_tree(x) for x in random.choice(grammar[symbol])]}

print(generate())

print(generate_tree())
